'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const isObject = (x) => typeof x === 'object' && x !== null;

class Branch {
    proxy; // Target object wrapped in proxy
    target; // Target object
    proxyTree; // ProxyTree the Branch belongs to
    childBranches = new Set([]); // Child Branches of this Branch
    /**
     * @public
     * Branch - Represents a target object wrapped in a Proxy().
     * In doing so, it keeps track of all its properties
     * and transforms them into a sub Branch as soon as they are accessed.
     * The Branch isn't aware of not accessed properties, as they aren't yet relevant.
     * @param proxyTree - Proxy Tree the Branch belongs to
     * @param target - Target Object the Branch represents and wraps a Proxy() around
     */
    constructor(proxyTree, target) {
        this.proxyTree = proxyTree;
        this.target = target;
        // Create Handler Methods which will be called by the Proxy() in order to intercepts specific Events
        const handler = {
            // get() handler intercepts an attempt to access a property of the target object (object this branch represents)
            get: (target, key) => {
                // If array.length, don't track it
                if (Array.isArray(target) && key === 'length')
                    return target.length;
                // If key doesn't exists in target, don't track it
                if (!(key in target))
                    return undefined;
                // If object has no own property at key, don't track it
                // This does the same as (key in target) but also filters method calls like (object.toString(), array.slice(), array.filter(), ..)
                if (!Object.prototype.hasOwnProperty.call(target, key))
                    return target[key];
                return this.recordUsage(target, key);
            },
        };
        // Wrap Proxy around target object
        // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Proxy
        this.proxy = new Proxy(target, handler);
    }
    /**
     * @internal
     * Record usage of an accessed property in the passed target object.
     * @param target - Target object in which a property at key was accessed
     * @param key - Key that was accessed in the target object
     */
    recordUsage(target, key) {
        if (key in target) {
            const value = target[key];
            // Check if property was accessed before and therefore has an existing Route to a sub Branch or value
            const branchRoute = this.getBranchRouteAtKey(key);
            // If Branch exists, increment the 'timesAccessed' property in order to keep track of how often it was accessed
            if (branchRoute) {
                branchRoute.timesAccessed += 1;
                return branchRoute.branch?.proxy || target[key];
            }
            // If Branch doesn't exist, create new Route and a sub Branch if no final primitive value got reached
            const newRoute = {
                key,
                timesAccessed: 1,
                branch: (isObject(value) &&
                    this.proxyTree.createBranch(value)) ||
                    null,
            };
            this.childBranches.add(newRoute);
            // Return proxyfied object of created Branch or the final primitive value if the Route doesn't lead to a sub Branch (object)
            return newRoute.branch?.proxy || target[key];
        }
        return undefined;
    }
    /**
     * @public
     * Checks if a Route to a particular property already exists in the sub Branches of this Branch.
     * @param key - Property key of the Route
     */
    getBranchRouteAtKey(key) {
        for (const branchRoute of this.childBranches) {
            if (branchRoute.key === key) {
                return branchRoute;
            }
        }
        return null;
    }
}

class ProxyTree$1 {
    rootBranch; // Root Branch of the proxy tree
    proxy; // Target object wrapped in proxy
    /**
     * @public
     * ProxyTree - Class that wraps around a target object, and its nested objects as you access them
     * in order to keep track of which properties were accessed via get/has proxy handlers.
     * @param target - Target Object
     */
    constructor(target) {
        this.rootBranch = this.createBranch(target);
        this.proxy = (this.rootBranch?.proxy || null);
    }
    /**
     * @public
     * Creates a new Branch of the ProxyTree which represents the passed target object.
     * @param target - Target Object
     */
    createBranch(target) {
        if (!isObject(target)) {
            console.error("ProxyTree: The ProxyTree accepts only values from the type 'object' and 'array'! " +
                `The passed type was '${typeof target}'! ` +
                'Learn more here: https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Proxy');
            return null;
        }
        return new Branch(this, target);
    }
    /**
     * @public
     * Transforms Proxy Tree into an easily processable object.
     */
    transformTreeToBranchObject() {
        let rootBranchUses = 0;
        // Calculate root Branch uses
        this.rootBranch.childBranches.forEach((childBranch) => (rootBranchUses += childBranch.timesAccessed));
        // Create root BranchObject
        const rootBranch = {
            key: 'root',
            timesAccessed: rootBranchUses,
            branches: [],
        };
        // Method that walks deeper into the Proxy Tree
        const walk = (branch, currentBranchObject) => {
            const childBranches = branch.childBranches;
            // Check if Branch has any sub Branches
            if (childBranches.size > 0) {
                // Go through sub Branches and transform them to BranchObjects
                childBranches.forEach((branchRoute) => {
                    const newBranchObject = {
                        key: branchRoute.key,
                        timesAccessed: branchRoute.timesAccessed,
                        branches: [],
                    };
                    // Add Sub Branch to the parent Branch ('branches' array)
                    currentBranchObject.branches.push(newBranchObject);
                    // Check if Route leads to any sub Branch.
                    // If so the Tree doesn't end here (-> sub object).
                    // So walk deeper and transform the deeper Branches into BranchObjects.
                    // Otherwise the Tree ends here and the Route leads to a primitive value like a number.
                    if (branchRoute.branch) {
                        walk(branchRoute.branch, newBranchObject);
                    }
                });
            }
        };
        // Start walking through the Proxy Tree
        walk(this.rootBranch, rootBranch);
        return rootBranch;
    }
    /**
     * @public
     * Returns the Paths to the accessed properties in array shape.
     * For example, an object `{ a: [{ b: 'c' }, { 1000: 'value' }, 'b'] }`,
     * has got the following paths pointing to existing properties:
     *
     * - `[]`
     * - `['a']`
     * - `['a', 0]`
     * - `['a', 0, 'b']`
     * - `['a', 1]`
     * - `['a', 1, 1000]`
     * - `['a', 2]`
     *
     * Be aware that this path points are only tracked if the accordingly property was actually accessed.
     * The Proxy Tree isn't aware of not accessed properties and thereby doesn't know the path to them
     * as they aren't relevant yet.
     */
    getUsedRoutes() {
        const usedRoutes = [];
        // Checks if path/route already exists in 'usedRoutes'
        const inRoutes = (path) => {
            for (const route of usedRoutes) {
                if (JSON.stringify(route) === JSON.stringify(path))
                    return true;
            }
            return false;
        };
        // Transform Proxy Tree into simple accessible object
        const rootBranchObject = this.transformTreeToBranchObject();
        const walk = (branchObject, path) => {
            // Check if Branch Children where accessed.
            // Otherwise here is an end, because whatever is behind this Branch got already tracked or was never accessed
            let branchChildRoutesTimesUsed = 0;
            branchObject.branches.forEach((childBranch) => (branchChildRoutesTimesUsed += childBranch.timesAccessed));
            // Check if Branch has any sub Branches and got accessed.
            // If so walk deeper into the Proxy Tree.
            // Otherwise push the path into the 'usedRoutes' array because this particular Path/Route ends here
            if (branchObject.branches.length > 0 &&
                branchObject.timesAccessed > 0 &&
                branchChildRoutesTimesUsed > 0) {
                // Go through sub Branches and walk into them if they got accessed
                branchObject.branches.forEach((branchChildObject) => {
                    if (branchChildObject.timesAccessed > 0) {
                        walk(branchChildObject, path ? [...path, branchChildObject.key] : [branchChildObject.key]);
                    }
                    // Decrease times accessed (because we passed this Branch in order to get to the child Branch)
                    branchObject.timesAccessed -= 1;
                });
            }
            else {
                // Push discovered Path into 'usedRoutes'
                if (path && !inRoutes(path))
                    usedRoutes.push(path);
                // Decrease times accessed
                if (branchObject.timesAccessed > 0)
                    branchObject.timesAccessed -= 1;
            }
        };
        // Start walking through the Proxy Tree.
        // Therefore walk through the Tree until the root Branch property 'timesAccessed' = 0.
        // Everytime it passes a Branch it decreases the 'timesAccessed' property.
        // This way it is able to reconstruct each used route to the acceded properties.
        while (rootBranchObject.timesAccessed > 0) {
            walk(rootBranchObject);
        }
        return usedRoutes;
    }
}

var ProxyTree = ProxyTree$1;

exports.Branch = Branch;
exports.ProxyTree = ProxyTree$1;
exports['default'] = ProxyTree;
